#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/errno.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/types.h>
#include <linux/net.h>

#include <net/sock.h>
#include <net/ip.h>

MODULE_LICENSE("GPL");

#define NETDISK_MAJOR 0
#define NETDISK_SECTORS 2880

#define KERNEL_SECTOR_SIZE 512
#define SERVER "10.1.1.1"
#define BOUNDARY "--netdisk1234"

static struct request_queue *netdisk_queue;
static unsigned char address[4]={ 10, 1, 1 , 1 };

static struct request_queue *queue;

static struct netdisk_device
{
  spinlock_t lock;
  struct gendisk *gd;
} NetDisk;

static void post_sector(unsigned char *sector, int block)
{
struct socket *sock;
struct sockaddr_in sin;
struct msghdr msg;
struct iovec iov;
char buffer[4096];
int error,len,r;
mm_segment_t oldfs;

  printk("NetDisk: post_sector(block=%d)\n",block);

  error=sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
  if (error<0)
  {
    printk("NetDisk: Couldn't create socket.\n");
    return;
  }

  sin.sin_family=AF_INET;
  // replace with inet_pton
  // sin.sin_addr.s_addr=inet_addr(SERVER);
  // inet_pton(AF_INET, SERVER, &sin.sin_addr.s_addr);
  memcpy(&sin.sin_addr.s_addr, address, 4);
  sin.sin_port=htons(80);

  error=sock->ops->connect(sock, (struct sockaddr *)&sin,sizeof(sin), 0);
  if (error<0)
  {
    printk("NetDisk: Couldn't connect.\n");
    return;
  }

  sprintf(buffer,"POST /netdisk/netdisk.php?op=write&block=%d HTTP/1.1\r\n"
                 "Host: %d.%d.%d.%d\r\n"
                 "Connection: close\r\n"
                 "Content-Type: multipart/form-data; boundary="BOUNDARY"\r\n"
                 "Content-Length: 659\r\n\r\n--"BOUNDARY"\r\n"
                 "Content-Disposition: form-data; name=\"sector_data\"; filename=\"null\"\r\n"
                 "Content-Type: application/octet-stream\r\n\r\n"
                 ,block,address[0],address[1],address[2],address[3]);
  len=strlen(buffer);
  memcpy(buffer+len,sector,512);
  sprintf(buffer+len+512,"\r\n--"BOUNDARY"--\r\n");

/*
  printk("Netdisk: len=%d content_length=%d\n",len,512+sizeof("\r\n--"BOUNDARY"--\r\n")-1+sizeof("--"BOUNDARY"\r\n""Content-Disposition: form-data; name=\"sector_data\"; filename=\"null\"\r\n"
                 "Content-Type: application/octet-stream\r\n\r\n"));
*/

  msg.msg_name=0;
  msg.msg_namelen=0;
  msg.msg_iov=&iov;
  msg.msg_iovlen=1;
  msg.msg_control=NULL;
  msg.msg_controllen=0;
  msg.msg_flags=MSG_DONTWAIT|MSG_NOSIGNAL;
  iov.iov_base=buffer;
  iov.iov_len=sizeof(buffer);

  oldfs=get_fs();
#ifndef __LCLINT__
  set_fs(KERNEL_DS);
#endif
  sock_sendmsg(sock, &msg, len+512+sizeof("\r\n--"BOUNDARY"--\r\n")-1);
  set_fs(oldfs);

  // while(len<512)
  {
    msg.msg_name=0;
    msg.msg_namelen=0;
    msg.msg_iov=&iov;
    msg.msg_iovlen=1;
    msg.msg_control=NULL;
    msg.msg_controllen=0;
    msg.msg_flags=MSG_DONTWAIT|MSG_NOSIGNAL;
    iov.iov_base=buffer;
    iov.iov_len=sizeof(buffer);

    oldfs=get_fs();
#ifndef __LCLINT__
    set_fs(KERNEL_DS);
#endif
    r=sock_recvmsg(sock, &msg, sizeof(buffer), 0);
    set_fs(oldfs);

  }

  printk("NetDisk: recv'd %d bytes\n",len);

  sock_release(sock);
}

static void get_sector(unsigned char *sector, int block)
{
struct socket *sock;
struct sockaddr_in sin;
struct msghdr msg;
struct iovec iov;
char buffer[4096];
int error,len,r;
mm_segment_t oldfs;
int flag;

  printk("NetDisk: get_sector(block=%d)\n",block);

  error=sock_create(PF_INET, SOCK_STREAM, IPPROTO_TCP, &sock);
  if (error<0)
  {
    printk("NetDisk: Couldn't create socket.\n");
    return;
  }

  printk("NetDisk: Going to connect()\n");

  memset(&sin,0,sizeof(sin));
  sin.sin_family=AF_INET;
  // replace with inet_pton
  // sin.sin_addr.s_addr=inet_addr(SERVER);
  // inet_pton(AF_INET, SERVER, &sin.sin_addr.s_addr);
  memcpy(&sin.sin_addr.s_addr, address, 4);
  // sin.sin_addr.s_addr=htonl(create_address(192,168,1,18));
  sin.sin_port=htons(80);

  error=sock->ops->connect(sock, (struct sockaddr *)&sin,sizeof(sin), O_RDWR);
  if (error<0)
  {
    printk("NetDisk: Couldn't connect.\n");
    return;
  }

  printk("NetDisk: Going to create request\n");

  sprintf(buffer,"GET /netdisk/netdisk.php?op=read&block=%d HTTP/1.1\r\nHost: %d.%d.%d.%d\r\nConnection: close\r\n\r\n",block,address[0],address[1],address[2],address[3]);

  msg.msg_name=0;
  msg.msg_namelen=0;
  msg.msg_iov=&iov;
  msg.msg_iovlen=1;
  msg.msg_control=NULL;
  msg.msg_controllen=0;
  msg.msg_flags=MSG_DONTWAIT|MSG_NOSIGNAL;
  iov.iov_base=buffer;
  iov.iov_len=sizeof(buffer);

  printk("NetDisk: Going to sock_sendmsg()\n");

  oldfs=get_fs();
#ifndef __LCLINT__
  set_fs(KERNEL_DS);
#endif

  sock_sendmsg(sock, &msg, strlen(buffer));
  set_fs(oldfs);

  printk("NetDisk: Going to sock_recvmsg()\n");

  len=0;
  while(len<4096)
  {
    msg.msg_name=0;
    msg.msg_namelen=0;
    msg.msg_iov=&iov;
    msg.msg_iovlen=1;
    msg.msg_control=NULL;
    msg.msg_controllen=0;
    msg.msg_flags=MSG_DONTWAIT|MSG_NOSIGNAL;
    iov.iov_base=buffer+len;
    iov.iov_len=sizeof(buffer);

    oldfs=get_fs();
#ifndef __LCLINT__
    set_fs(KERNEL_DS);
#endif
    r=sock_recvmsg(sock, &msg, 4096-len, 0);
    set_fs(oldfs);

    printk("sock_recvmsg()=%d\n",r);
    if (r<=0) break;
    len=len+r;

  }

  printk("NetDisk: recv'd %d bytes\n",len);

  r=0;
  flag=0;
  while(r<=len-512)
  {
    if (buffer[r]=='\r')
    {
    }
      else
    if (buffer[r]=='\n')
    {
      flag++;
      if (flag==2)
      {
/*
        buffer[r]=0;
        printk("%s",buffer);
*/
        r=r+1;
        if (len-r>=512)
        {
          memcpy(sector, buffer+r, 512);
        }
        break;
      }
    }
      else
    { flag=0; }
    r++;
  }

/*
  buffer[r]=0;
  printk("r=%d len=%d \"%s\"\n",r,len,buffer);
*/

  printk("NetDisk: release\n");

  sock_release(sock);

  printk("NetDisk: get_sector() exit \n");
}

static void netdisk_request(struct request_queue *queue)
{
struct request *req;

  printk("NetDisk: request()\n");

  req=elv_next_request(queue);

  while ((req=elv_next_request(queue))!=NULL)
  {
    if (!blk_fs_request(req))
    {
      end_request(req, 0);
      continue;
    }

    // Send data to server
    if (rq_data_dir(req))  // write block
    {
      post_sector(req->buffer, req->sector);
    }
      else // read block
    {
      get_sector(req->buffer, req->sector);
    }

    end_request(req, 1);
  }

}

static int netdisk_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
struct hd_geometry geo;

  printk("NetDisk: ioctl(%d) %ld\n",cmd,arg);

  if (cmd==HDIO_GETGEO)
  {
    geo.cylinders=1474560>>6;  // dunno :(
    geo.heads=2;
    geo.sectors=18;
    geo.start=0;
    if (copy_to_user((void *)arg, &geo, sizeof(geo))) return -EFAULT;

    return 0;
  }

  return -ENOTTY;
}

static struct block_device_operations netdisk_fops =
{
  .owner = THIS_MODULE,
  // .read = netdisk_read,
  // .write = netdisk_write,
  // .open = netdisk_open,
  // .release = netdisk_release
  .ioctl = netdisk_ioctl
};

static int __init netdisk_init(void)
{
int major_num;

  printk("NetDisk: init()\n");

  spin_lock_init(&NetDisk.lock);

  printk("NetDisk: blk_init_queue()\n");
  queue = blk_init_queue(netdisk_request, &NetDisk.lock);
  if (queue == NULL)
  {
    printk("NetDisk: queue is NULL\n");
    // FIXME - do something about it
  }

  printk("NetDisk: blk_queue_hardsect_size()\n");
  blk_queue_hardsect_size(queue, 512);

  printk("NetDisk: register_blkdev()\n");
  major_num=register_blkdev(NETDISK_MAJOR,"netdisk");

  if (major_num<=0)
  {
    printk(KERN_WARNING "NetDisk: Couldn't register block device\n");
    return -1;
  }

  printk("NetDisk: alloc_disk()\n");
  NetDisk.gd=alloc_disk(1);
  if (!NetDisk.gd)
  {
    printk("Netdisk: Error, could not alloc_disk()\n");
    // FIXME - put something here
    return -1;
  }

  printk("NetDisk: set up gd)\n");
  NetDisk.gd->major=major_num;
  NetDisk.gd->first_minor=0;
  NetDisk.gd->fops=&netdisk_fops;
  NetDisk.gd->private_data=&NetDisk;   // not really needed right now
  strcpy(NetDisk.gd->disk_name, "netdisk0");
  set_capacity(NetDisk.gd, NETDISK_SECTORS*(512/KERNEL_SECTOR_SIZE));
  NetDisk.gd->queue=queue;
  add_disk(NetDisk.gd);

  printk("Netdisk: init() sucessful\n");

  return 0;
}

static void __exit netdisk_exit(void)
{
int major_num;

  printk("NetDisk: exit()\n");

  major_num=NetDisk.gd->major;

  del_gendisk(NetDisk.gd);
  put_disk(NetDisk.gd);
  unregister_blkdev(major_num, "netdisk");
  blk_cleanup_queue(queue);
}

/*
int netdisk_open(struct inode *, struct file *)
{
  printk("NetDisk: open()\n");

}
*/

module_init(netdisk_init);
module_exit(netdisk_exit);



